import numpy as np
from rombus.model import RombusModel
from typing import NamedTuple
import math

# This sample code provides a starting point for building your own Reduced Order Models
# Edit it to suit your needs.  In its initial form, it provides a model for building
# a reduced order model (ROM) of the function sinc(x)=sin(x)/x

# The following class is the essential element needed to define a model for RombusModel
# It must inherit from the RombusModel Class (which needs to be imported above).  The class
# must include one-or-more calls to 'params.add_param()' to define the names and limits of
# the parameters of the model and define the following methods:
#
#   1) set_domain(): sets the domain over-and-on which the ROM will be defined
#
#   2) compute(): returns a model for a given parameter set for each value of domain
#
class model(RombusModel):
    """Class for creating a ROM for the function sinc(x)=sin(x)/x"""

    # This sets the datatype generated by the model (default: 'float64')
    model_dtype = np.dtype("float64")

    # Add as many calls to params.add() as your model has parameters.
    # Samples will be sent to the 'compute()' method (see below) as
    # named tuples and will be addressable using the names you give them.
    #
    # Note that you may need to silence linting errors by adding a '# noqa F821'
    # directive at the end.  params is initialised, it's just done in a way that
    # IDEs and linters have troubles following.
    #
    # Syntax: params.add(name, min_value_allowed,max_value_allowed)
    params.add("A", 0, 10)  # noqa F821

    # Anything added as a class member here will just be computed once and will
    # be accessible in the 'compute()' and 'set-domain()' methods below.
    def cache(self):
        self.x_min = 0
        self.x_max = 100
        self.n_x = 1024

    # Set the domain over-and-on which the ROM will be defined
    def set_domain(self) -> np.array:
        return np.linspace(self.x_min, self.x_max, self.n_x)

    # Compute the model.  This function should accept a named tuple with
    # member names given by the params.add() calls above and iter should
    # return a numpy array of type given my 'model_dtype' given above
    def compute(self, params: NamedTuple, x) -> np.array:
        """Compute the model for a given parameter set."""

        return sinc_vectorized(params.A * x)


# Create the function that will compute our model
def sinc_scalar(x):
    """A scalar version of our model function, which takes a
    value x and returns a value x."""

    if x == 0.0:
        return 1.0
    else:
        return math.sin(x) / x


# Use np.vectorize() to create a function that accepts a
# numpy array and returns a numpy array
sinc_vectorized = np.vectorize(sinc_scalar)



import numpy as np
import importlib
import pytest
from click.testing import CliRunner

from rombus.cli import cli


@pytest.mark.lalsuite
def test_PhenomP(tmp_path):

    atol = 1e-12
    rtol = 1e-12

    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):

        greedy_filename = str(
            importlib.resources.files("rombus.models").joinpath("PhenomP_samples.csv")
        )
        result = runner.invoke(
            cli,
            [
                "build",
                "rombus.models.PhenomP:model",
                greedy_filename,
            ],
        )
        assert result.exit_code == 0
        result = runner.invoke(
            cli,
            [
                "evaluate",
                "PhenomP.hdf5",
                "m1=23.07487209351506",
                "m2=29.753779794217984",
                "chi1L=0.3803919479347709",
                "chi2L=0.266269755913278",
                "chip=0.5458029422691579",
                "thetaJ=1.7621026471916568",
                "alpha=2.765791991075064",
            ],
        )
        assert result.exit_code == 0
        delta_diff = np.load("ROM_diff.npy")

    assert np.allclose(delta_diff, 0.0, rtol=rtol, atol=atol)



# --- integrals.py ---

"""
    Classes and functions for computing inner products of functions
"""

import numpy as np

from .training import TrainingTools


def rate_to_num(a, b, rate):
    """Convert sample rate to sample numbers in [a,b]"""
    return np.floor(float(b - a) * rate) + 1


def num_to_rate(a, b, num):
    """Convert sample numbers in [a,b] to sample rate"""
    return (num - 1.0) / np.float(b - a)


##############################################
class Quadratures(TrainingTools):
    def __init__(self):
        TrainingTools.__init__(self)
        self.switch_dict = {
            "riemann-rate": self.riemann_rate,
            "riemann-num": self.riemann_num,
            "trapezoidal-rate": self.trapezoidal_rate,
            "trapezoidal-num": self.trapezoidal_num,
            "chebyshev": self.chebyshev,
            "chebyshev-gauss-lobatto": self.chebyshev_gauss_lobatto,
            "legendre": self.legendre,
            "legendre-gauss-lobatto": self.legendre_gauss_lobatto,
            # 'roq': self.roq
        }
        self.options = self.switch_dict.keys()

    def __call__(self, args, quad_type):  # , multidomainQ=False):

        dim = np.shape(args[:-1])
        num_or_rate = args[-1]

        # Check that the number of domains equals to the
        #   number of samples/sample rates specified
        if type(num_or_rate) == list:
            if len(num_or_rate) != dim[0] - 1:
                raise Exception(
                    (
                        "Expecting the numbers of intervals and "
                        "samples/sample rates to be equal."
                    )
                )

        # Check how many quadrature rules are specified
        if type(quad_type) == list:
            type_quad_type = "list"
            if len(quad_type) != len(num_or_rate):
                raise Exception(
                    (
                        "Expecting the numbers of intervals and "
                        "specified quadrature rules to be equal."
                    )
                )
        else:
            type_quad_type = "str"

        # Make quadrature rules for single or multiple domains
        if len(dim) == 1:

            # Single integration domain
            if dim[0] == 2:
                ans = self.switch_dict[quad_type](*args)

            # Multiple integration domains
            if dim[0] > 2:
                if type_quad_type == "str":
                    ans = np.hstack(
                        [
                            self.switch_dict[quad_type](
                                args[ii], args[ii + 1], num_or_rate[ii]
                            )
                            for ii in range(dim[0] - 1)
                        ]
                    )
                elif type_quad_type == "list":
                    ans = np.hstack(
                        [
                            self.switch_dict[quad_type[ii]](
                                args[ii], args[ii + 1], num_or_rate[ii]
                            )
                            for ii in range(dim[0] - 1)
                        ]
                    )

        elif len(dim) > 1:
            # CHECKME
            # For integrating in multiple dimensions
            nodes, weights = [], []
            for ii in range(len(args[0])):
                arg = np.transpose(args)[ii]
                quad = quad_type[ii]
                nws = self.switch_dict[quad](*arg)
                nodes.append(nws[0])
                weights.append(nws[1])
            weights_nd = np.transpose(self.tuple_to_vstack(self.meshgrid2(*weights)))
            ans = [
                self.tuple_to_vstack(self.meshgrid2(*nodes)).T,
                np.array(
                    [np.prod(weights_nd[ii]) for ii in range(len(weights_nd[:, 0]))]
                ),
            ]

        return ans

    def riemann_num(self, a, b, n):
        """
        Uniformly sampled array using Riemann quadrature rule
        over given interval with given number of samples

        Input
        -----
        a -- start of interval
        b -- end of interval
        n -- number of quadrature points

        Output
        ------
        nodes   -- quadrature nodes
        weights -- quadrature weights
        """
        nodes = np.linspace(a, b, num=n)
        weights = np.ones(n, dtype="double")
        return [nodes, (b - a) / (n - 1.0) * weights]

    def riemann_rate(self, a, b, rate):
        """
        Uniformly sampled array using Riemann quadrature rule
        over given interval with given sample rate

        Input
        -----
        a    -- start of interval
        b    -- end of interval
        rate -- sample rate

        Output
        ------
        nodes   -- quadrature nodes
        weights -- quadrature weights
        """
        n = rate_to_num(a, b, rate)
        return self.riemann_num(a, b, n)

    def trapezoidal_num(self, a, b, n):
        """
        Uniformly sampled array using trapezoidal quadrature rule
        over given interval with given number of samples

        Input
        -----
        a -- start of interval
        b -- end of interval
        n -- number of quadrature samples

        Output
        ------
        nodes   -- quadrature nodes
        weights -- quadrature weights
        """
        nodes = np.linspace(a, b, num=n)
        weights = np.ones(n, dtype="double")
        weights[0] /= 2.0
        weights[-1] /= 2.0
        return [nodes, weights * (b - a) / (n - 1.0)]

    def trapezoidal_rate(self, a, b, rate):
        """
        Uniformly sampled array using trapezoidal quadrature rule
        over given interval with given sample rate

        Input
        -----
        a    -- start of interval
        b    -- end of interval
        rate -- sample rate

        Output
        ------
        nodes   -- quadrature nodes
        weights -- quadrature weights
        """
        n = rate_to_num(a, b, rate)
        return self.trapezoidal_num(a, b, n)

    def chebyshev(self, a, b, n):
        """
        Uniformly sampled array using Chebyshev-Gauss quadrature rule
        over given interval with given number of samples

        Input
        -----
        a -- start of interval
        b -- end of interval
        n -- number of quadrature samples

        Output
        ------
        nodes   -- quadrature nodes
        weights -- quadrature weights
        """
        num = int(n) - 1.0
        nodes = np.array(
            [
                -np.cos(np.pi * (2.0 * ii + 1.0) / (2.0 * num + 2.0))
                for ii in range(int(n))
            ]
        )
        weights = np.pi / (num + 1.0) * np.sqrt(1.0 - nodes**2)
        return [nodes * (b - a) / 2.0 + (b + a) / 2.0, weights * (b - a) / 2.0]

    def chebyshev_gauss_lobatto(self, a, b, n):
        """
        Uniformly sampled array using Chebyshev-Gauss-Lobatto quadrature rule
        over given interval with given number of samples

        Input
        -----
        a -- start of interval
        b -- end of interval
        n -- number of quadrature samples

        Output
        ------
        nodes   -- quadrature nodes
        weights -- quadrature weights
        """
        num = int(n) - 1.0
        nodes = np.array([-np.cos(np.pi * ii / num) for ii in range(int(n))])
        weights = np.pi / num * np.sqrt(1.0 - nodes**2.0)
        weights[0] /= 2.0
        weights[-1] /= 2.0
        return [nodes * (b - a) / 2.0 + (b + a) / 2.0, weights * (b - a) / 2.0]

    def legendre(self, a, b, num):
        raise Exception("Legendre Gauss quadrature rule is not yet implemented.")
        pass

    def legendre_gauss_lobatto(self, a, b, num):
        raise Exception(
            "Legendre Gauss-Lobatto quadrature rule is not yet implemented."
        )
        pass


#     def roq(self, x, w, eim_indices, B):
#         """Reduced-order quadrature rule
#
#         Input
#         -----
#         x             -- original quadrature nodes (e.g., Chebyshev)
#         w             -- original quadrature weights
#         eim_indices    -- array indices of empirical interpolation nodes
#         B             -- empirical interpolant operator, `B`
#
#         Returns
#         -------
#         nodes     -- reduced-order quadrature nodes
#         weights -- reduced-order quadrature weights
#
#         Examples
#         --------
#         Let x, w be 20 Chebyshev Gauss nodes and weights over the interval
#         [10, 100] from
#
#           >>> qu = rp.Quadratures()
#
#           >>> x, w = qu([10, 100, 20], 'chebyshev')
#
#         If `B` is the empirical interpolant operator so that the empirical
#         interpolant acting on a function `h` is I[h] = \sum_{i=1}^m B_i*h_i
#         and `inds` are the array indices associated with the empirical
#         interpolation nodes then
#
#           >>> roq_x, roq_w = rp.Quadratures()(x, w, inds, B)
#
#         gives the reduced-order quadrature nodes (`roq_x`) and weights
#         (`roq_w`).
#
#         """
#
#         nodes = x[eim_indices]
#         weights = np.dot(B, w)
#         return [nodes, weights]


##############################################
class Real:
    def __init__(self, args, quad_type="trapezoidal-num"):
        self.quad_type = quad_type
        self.nodes, self.weights = Quadratures()(args, quad_type)

    def integral(self, h):
        """Integral of a function"""
        return np.dot(self.weights, np.real(h))

    def dot(self, h1, h2):
        """Dot product of two functions"""
        dim = np.shape(h1)
        h1real = np.real(h1)
        h2real = np.real(h2)
        if len(dim) == 1:
            return np.dot(self.weights, h1real * h2real)
        else:
            return np.dot(
                self.weights, np.sum(h1real[ii] * h2real[ii] for ii in range(dim[0]))
            )

    def norm(self, h):
        """Norm of function"""
        hreal = np.real(h)
        return np.sqrt(np.real(self.dot(hreal, hreal)))

    def normalize(self, h):
        """Normalize a function"""
        return h / self.norm(h)

    def match(self, h1, h2):
        """Match integral"""
        h1n = self.normalize(h1)
        h2n = self.normalize(h2)
        return np.real(self.dot(h1n, h2n))

    def mismatch(self, h1, h2):
        """Mismatch integral (1-match)"""
        return 1.0 - self.match(h1, h2)


##############################################
class Complex:
    def __init__(self, args, quad_type="trapezoidal-num"):
        self.quad_type = quad_type
        self.nodes, self.weights = Quadratures()(args, quad_type)

    def integral(self, h):
        """Integral of a function"""
        return np.dot(self.weights, h)

    def dot(self, h1, h2):
        """Dot product of two functions"""
        dim = np.shape(h1)
        if len(dim) == 1:
            return self.integral(np.conj(h1) * h2)
        else:
            return self.integral(
                np.sum(np.conj(h1[ii]) * h2[ii] for ii in range(dim[0]))
            )

    def norm(self, h):
        """Norm of function"""
        return np.sqrt(np.real(self.dot(h, h)))

    def normalize(self, h):
        """Normalize a function"""
        return h / self.norm(h)

    def match(self, h1, h2):
        """Match integral"""
        h1n = self.normalize(h1)
        h2n = self.normalize(h2)
        return np.real(self.dot(h1n, h2n))

    def mismatch(self, h1, h2):
        """Mismatch integral (1-match)"""
        return 1.0 - self.match(h1, h2)


##############################################
class InnerProduct(Real, Complex):
    """Integrals for computing inner products and norms of functions"""

    def __init__(self, args, quad_type="trapezoidal-num"):

        Real.__init__(self, args, quad_type)
        self.real = Real(args, quad_type)

        Complex.__init__(self, args, quad_type)
        self.complex = Complex(args, quad_type)

        self.switch_dict = {
            "complex": self.complex,
            "real": self.real,
        }
        self.options = self.switch_dict.keys()

    def __call__(self, h1, h2, inner_type):
        return self.dot(h1, h2, inner_type)

    def integral(self, h, inner_type):
        """Integral of a function"""
        return self.switch_dict[inner_type].integral(h)

    def dot(self, h1, h2, inner_type):
        """Dot product of two functions"""
        return self.switch_dict[inner_type].dot(h1, h2)

    def norm(self, h, inner_type):
        """Norm of function"""
        return np.sqrt(np.real(self.dot(h, h, inner_type)))

    def normalize(self, h, inner_type):
        """Normalize a function"""
        return h / self.norm(h, inner_type)

    def match(self, h1, h2, inner_type):
        """Match integral"""
        return self.switch_dict[inner_type].match(h1, h2)

    def mismatch(self, h1, h2, inner_type):
        """Mismatch integral (1-match)"""
        return self.switch_dict[inner_type].mismatch(h1, h2)

    def Linfty(self, h):
        """L-infinity norm"""
        # FIXME: h cannot be a multi-dim array
        return np.max(np.abs(h))

    def Ln(self, h, n):
        """L-n norm"""
        # FIXME: h cannot be a multi-dim array
        if n > 0:
            return (np.dot(self.weights, np.abs(h) ** n)) ** (1.0 / n)
        pass

    def L2(self, h):
        """L-2 norm"""
        # FIXME: h cannot be a multi-dim array
        return self.Ln(h, 2)

    def _test_monomial(self, n=0):
        """Test integration rule by integrating the monomial x**n"""
        ans = self.integral(self.nodes**n, "real")
        a, b = self.nodes[0], self.nodes[-1]
        expd = (b ** (n + 1) - a ** (n + 1)) / (n + 1)

        print("\nExpected value for integral =", expd)
        print("Computed value for integral =", ans)
        print("\nAbsolute difference =", expd - ans)
        print("Relative difference =", 1.0 - ans / expd)

        pass



import numpy as np


def dot_product(weights, a, b):

    assert len(a) == len(b)
    return np.vdot(a * weights, b)


def project_onto_basis(integration_weights, e, my_ts, iter, dtype):

    # c = np.einsum('i,ji->j', integration_weights*e[iter].conjugate(), h)
    pc = np.zeros(len(my_ts), dtype=dtype)
    # projections += np.outer(c,e[iter])
    for j in range(len(my_ts)):
        pc[j] = dot_product(integration_weights, e[iter], my_ts[j])
        # projections[j] += pc[j][iter]*e[iter]
    return pc


def MGS(RB, next_vec, iter):
    """what is this doing?"""
    dim_RB = iter
    for i in range(dim_RB):
        # --- ortho_basis = ortho_basis - L2_proj*basis; ---
        L2 = np.vdot(RB[i], next_vec)
        next_vec -= RB[i] * L2
    norm = np.sqrt(np.vdot(next_vec, next_vec))
    next_vec /= norm
    return next_vec, norm


def IMGS(RB, next_vec, iter):
    """what is this doing?"""
    ortho_condition = 0.5
    norm_prev = np.sqrt(np.vdot(next_vec, next_vec))
    flag = False
    while not flag:
        next_vec, norm_current = MGS(RB, next_vec, iter)
        next_vec *= norm_current
        if norm_current / norm_prev <= ortho_condition:
            norm_prev = norm_current
        else:
            flag = True
        norm_current = np.sqrt(np.vdot(next_vec, next_vec))
        next_vec /= norm_current
    # RB[iter] = next_vec  # np.vstack((RB, next_vec))
    # return RB[iter]
    return next_vec


def get_highest_error(error_list):
    rank, idx, err = -np.inf, -np.inf, -np.inf
    for rank_id, rank_errors in enumerate(error_list):
        max_rank_err = max(rank_errors)
        if max_rank_err > err:
            err = max_rank_err
            idx = rank_errors.index(err)
            rank = rank_id
    return rank, idx, np.float64(err.real)



# --- fit.py ---

"""
    Classes and functions for fitting functions at the greedy-selected parameter values.
"""


import numpy as np
from scipy.interpolate import splev, splrep

# from scipy.optimize import curve_fit
from scipy.optimize import leastsq

from . import lib

##############################################
# The following three functions (_general_function, _weighted_general_function,
# curve_fit) are from scipy.optimize. I have changed the error handling here
# so that curve_fit returns the parameters found, even if they are not optimal.
# The original code raises an exception and aborts the program. Using a `try`
# and `except` pair does not seem to retain the (suboptimal) fit parameters.


def _general_function(params, xdata, ydata, function):
    return function(xdata, *params) - ydata


def _weighted_general_function(params, xdata, ydata, function, weights):
    return weights * (function(xdata, *params) - ydata)


def curve_fit(f, xdata, ydata, p0=None, sigma=None, **kw):
    """
    Use non-linear least squares to fit a function, f, to data.

    Assumes ``ydata = f(xdata, *params) + eps``

    Parameters
    ----------
    f : callable
        The model function, f(x, ...).  It must take the independent
        variable as the first argument and the parameters to fit as
        separate remaining arguments.
    xdata : An N-length sequence or an (k,N)-shaped array
        for functions with k predictors.
        The independent variable where the data is measured.
    ydata : N-length sequence
        The dependent data --- nominally f(xdata, ...)
    p0 : None, scalar, or M-length sequence
        Initial guess for the parameters.  If None, then the initial
        values will all be 1 (if the number of parameters for the function
        can be determined using introspection, otherwise a ValueError
        is raised).
    sigma : None or N-length sequence
        If not None, this vector will be used as relative weights in the
        least-squares problem.

    Returns
    -------
    popt : array
        Optimal values for the parameters so that the sum of the squared error
        of ``f(xdata, *popt) - ydata`` is minimized
    pcov : 2d array
        The estimated covariance of popt.  The diagonals provide the variance
        of the parameter estimate.

    See Also
    --------
    leastsq

    Notes
    -----
    The algorithm uses the Levenberg-Marquardt algorithm through `leastsq`.
    Additional keyword arguments are passed directly to that algorithm.

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.optimize import curve_fit
    >>> def func(x, a, b, c):
    ...     return a*np.exp(-b*x) + c

    >>> x = np.linspace(0,4,50)
    >>> y = func(x, 2.5, 1.3, 0.5)
    >>> yn = y + 0.2*np.random.normal(size=len(x))

    >>> popt, pcov = curve_fit(func, x, yn)

    """
    if p0 is None:
        # determine number of parameters by inspecting the function
        import inspect

        args, varargs, varkw, defaults = inspect.getargspec(f)
        if len(args) < 2:
            msg = "Unable to determine number of fit parameters."
            raise ValueError(msg)
        if "self" in args:
            p0 = [1.0] * (len(args) - 2)
        else:
            p0 = [1.0] * (len(args) - 1)

    if np.isscalar(p0):
        p0 = np.array([p0])

    args = (xdata, ydata, f)
    if sigma is None:
        func = _general_function
    else:
        func = _weighted_general_function
        args += (1.0 / np.asarray(sigma),)

    # Remove full_output from kw, otherwise we're passing it in twice.
    return_full = kw.pop("full_output", False)
    res = leastsq(func, p0, args=args, full_output=1, **kw)
    (popt, pcov, infodict, errmsg, ier) = res

    if ier not in [1, 2, 3, 4]:
        msg = "Optimal parameters not found: " + errmsg
        # raise RuntimeError(msg)
        print(msg)
        return popt, pcov
        # if return_full:
        #    return popt, pcov, infodict, errmsg, ier
        # else:
        #    return popt, pcov

    if (len(ydata) > len(p0)) and pcov is not None:
        s_sq = (func(popt, *args) ** 2).sum() / (len(ydata) - len(p0))
        pcov = pcov * s_sq
    else:
        pcov = np.inf

    if return_full:
        return popt, pcov, infodict, errmsg, ier
    else:
        return popt, pcov


##############################################
class Spline1d:
    """Class for building one dimensional spline interpolants"""

    def __init__(self, x, y, k=3):
        """Initialize Spline1d class. Builds 1d spline interpolant
        of degree k from x and y data.

        Input
        -----
           x   -- given array of parameter values
           y   -- function values to be interpolated (e.g., at empirical
                           interpolation nodes
           k    -- degree of spline (default is 3)
        """

        args = np.argsort(x)  # `x` values must be sorted in ascending order
        self.dim = np.shape(y)
        self.degree = k
        if len(self.dim) == 1:
            self.knots, self.coeffs, _ = splrep(x[args], y[args], k=k)
        elif len(self.dim) > 1:
            self.coeffs = []
            for yy in y:
                temp = splrep(x[args], yy[args], k=k)
                self.coeffs.append(temp[1])
                self.knots = temp[0]
        else:
            raise Exception("Expected a non-empty array.")

        self.fitparams = [self.knots, self.coeffs, self.degree]

    def __call__(self, p, der=0):
        return self.fit(p, der)

    def fit(self, p, der=0):
        if len(self.dim) == 1:
            return splev(p, self.fitparams, der)
        elif len(self.dim) > 1:
            return np.array(
                [
                    splev(p, [self.knots, self.coeffs[ii], self.degree])
                    for ii in range(self.dim[0])
                ]
            )


##############################################
class Polynomial:
    def __init__(self, x, y, deg=3):

        # Map x to reference interval, [-1,1]
        ref = lib.map_intervals(x, -1.0, 1.0)

        self.shape = np.shape(y)
        if len(self.shape) == 1:
            self.fitparams_ref = np.polyfit(ref, y, deg=deg)
            self.fitparams = self.map_poly_coeffs(
                self.fitparams_ref, ref, np.min(x), np.max(x)
            )
        elif len(self.shape) > 1:
            if len(np.shape(deg)) == 0:
                self.fitparams_ref = [
                    np.polyfit(ref, y[ii], deg=deg) for ii in range(self.shape[0])
                ]
            else:
                self.fitparams_ref = [
                    np.polyfit(ref, y[ii], deg=deg[ii]) for ii in range(self.shape[0])
                ]
            self.fitparams = [
                self.map_poly_coeffs(pp, ref, np.min(x), np.max(x))
                for pp in self.fitparams_ref
            ]
        else:
            raise Exception("Expected a non-empty array.")
        pass

    def map_poly_coeffs(self, coeffs, x, a, b):
        """Map polynomial coefficients from p(x) to that on the interval [a,b]"""

        # Linear transformation from x to [a,b]
        M = (np.max(x) - np.min(x)) / (b - a)
        B = np.min(x) - M * a

        # Map the polynomial coefficients
        deg, mapped_coeffs = len(coeffs), []
        c = coeffs[::-1]  # Reverse order of polynomial coefficients
        for n in range(deg):
            mapped_coeffs.append(
                M**n
                * np.sum(
                    lib.choose(kk, n) * c[kk] * B ** (kk - n) for kk in range(n, deg)
                )
            )

        return mapped_coeffs[
            ::-1
        ]  # Reverse order of polynomial coefficients for use with np.polyval

    def __call__(self, p):
        return self.fit(p)

    def fit(self, p):
        if len(self.shape) == 1:
            return np.polyval(self.fitparams, p)
        elif len(self.shape) > 1:
            return np.array(
                [np.polyval(self.fitparams[ii], p) for ii in range(self.shape[0])]
            )


##############################################
class UserDefined:
    def __init__(self, x, y, fn, p0=None, maxfev=1000):
        self.shape = y.shape
        self.fn = fn
        if len(self.shape) == 1:
            self.fitparams = curve_fit(fn, x, y, p0=p0, maxfev=maxfev)[0]
        elif len(self.shape) > 1:
            self.fitparams = [
                curve_fit(fn, x, y[ii], p0=p0, maxfev=maxfev)[0]
                for ii in range(self.shape[0])
            ]
        else:
            raise Exception("Expected a non-empty array.")
        pass

    def __call__(self, p):
        return self.fit(p)

    def fit(self, p):
        if len(self.shape) == 1:
            return self.fn(p, *self.fitparams)
        elif len(self.shape) > 1:
            return np.array(
                [self.fn(p, *self.fitparams[ii]) for ii in range(self.shape[0])]
            )


##############################################
# class Surrogate(Spline1d, Polynomial, UserDefined, eim.EmpiricalInterpolation):
class Surrogate(Spline1d, Polynomial, UserDefined):
    def __init__(self, args, fit_type):
        # eim.EmpiricalInterpolation.__init__(self)

        if fit_type == "spline-1d":
            Spline1d.__init__(self, *args)
            self.fit = Spline1d(*args).fit
        elif fit_type == "poly":
            Polynomial.__init__(self, *args)
            self.fit = Polynomial(*args).fit
        elif fit_type == "userdef":
            UserDefined.__init__(self, *args)
            self.fit = UserDefined(*args).fit

    def __call__(self, p, B):
        return eval(p, B)

    def eval(self, p, B):
        return np.dot(self.fit(p), B)



import numpy as np
import pylab as plt


def errors(err_list):
    plt.plot(err_list)
    plt.xlabel("# Basis elements")
    plt.ylabel("Error")
    plt.yscale("log")
    plt.tight_layout()
    plt.savefig("basis_error.png")


def basis(rb_matrix):
    num_elements = len(rb_matrix)
    total_frames = 125
    n_models_per_frame = int(num_elements / total_frames)
    if n_models_per_frame < 1:
        n_models_per_frame = 1
    fig, ax = plt.subplots(total_frames, 1, figsize=(4.5, 2.5 * total_frames))
    for i in range(total_frames):
        start_i = int(i * n_models_per_frame)
        end_i = int(start_i + n_models_per_frame)
        for model_id in range(start_i, end_i):
            if end_i < num_elements:
                h = rb_matrix[model_id]
                ax[i].plot(h, color=f"C{model_id}", alpha=0.7)
        ax[i].set_title(f"Basis element {start_i:003}-{end_i:003}")
    plt.tight_layout()
    fig.savefig("basis.png")


def compare_rom_to_true(ROM, model_params_in):
    """Compare computed ROM to input model"""

    model_params = ROM.model.params.params_dtype(**model_params_in)

    domain = ROM.model.domain

    model_full = ROM.model.compute(model_params, domain)
    model_nodes = ROM.model.compute(model_params, ROM.empirical_interpolant.nodes)
    model_rom = ROM.evaluate(model_params)

    plt.semilogx(domain, model_rom, label="ROM", alpha=0.5, linestyle="--")
    plt.semilogx(domain, model_full, label="Full model", alpha=0.5)
    plt.scatter(ROM.empirical_interpolant.nodes, model_nodes, s=1)
    plt.legend()

    np.save("ROM_diff", np.subtract(model_rom, model_full))
    plt.savefig("comparison.pdf", bbox_inches="tight")



# --- __init__.py ---

"""Rombus
   =====

   Provides
     1. Functionality to build problem-specific greedy algorithms
     2. Reduced basis representations of function spaces
     3. Empirical interpolation methods
     4. Reduced-order quadrature rules
     5. Empirical Fourier transforms

   How to use the documentation
   -----------------------------
   Documentation is available via docstrings provided with the code.

   We recommend exploring the docstrings using
   `IPython <http://ipython.scipy.org>`, an advanced Python shell with
   TAB-completion and introspection capabilities. See below for further
   instructions.

   The docstring examples assume that `rombus` has been imported as `rb`::

     >>> import rombus as rb

   Code snippets are indicated by three greater-than signs.

   Use the built-in ``help`` function to view a function's or class's
   docstring::

     >>> help(rb.greedy)

   Available subpackages
   ---------------------
   algorithms
       Standard reduced-order modeling routines

   eim
       Core empirical interpolation tools

   greedy
       Core greedy algorithm tools

   integrals
       Tools for quadrature rules and integration

   lib
       Basic functions used by several subpackages

   training
       Core training set generation tools

   Viewing documentation using IPython
   -----------------------------------
   Start IPython and load Rombus (``import rombus as rb``), which
   will import `rombus` under the alias `rb`. Then, use the ``cpaste``
   command to paste examples into the shell. To see which functions
   are available in `rombus`, type ``rb.<TAB>`` (where ``<TAB>`` refers
   to the TAB key), or use ``rb.greedy?<ENTER>`` (where ``<ENTER>``
   refers to the ENTER key) to narrow down the list. To view the
   docstring for a function or class, use ``rb.greedy?<ENTER>`` (to
   view the docstring) and ``rb.greedy??<ENTER>`` (to view the source
   code).
   """



import sys
import timeit
from typing import List

import h5py
import mpi4py
import numpy as np

import rombus._core.algorithms as algorithms
import rombus._core.misc as misc
import rombus._core.mpi as mpi
from rombus.model import RombusModel

DEFAULT_TOLERANCE = 1e-14
DEFAULT_REFINE_N_RANDOM = 100


class ROM(object):
    def __init__(
        self,
        model,
        samples,
        reduced_basis=None,
        empirical_interpolant=None,
        tol=DEFAULT_TOLERANCE,
    ):

        if isinstance(model, str):
            self.model = RombusModel.load(model)
        elif isinstance(model, RombusModel):
            self.model = model
        else:
            raise Exception

        self.model = model
        self.samples = samples
        self.reduced_basis = reduced_basis
        self.empirical_interpolant = empirical_interpolant

    def build(self, do_step=None, tol=DEFAULT_TOLERANCE):

        if do_step is None or do_step == "RB":
            self.reduced_basis = ReducedBasis().compute(
                self.model, self.samples, tol=tol
            )

        if do_step is None or do_step == "EI":
            if self.reduced_basis is None:
                raise Exception
            self.empirical_interpolant = EmpiricalInterpolant().compute(
                self.reduced_basis
            )

        return self

    def refine(
        self, n_random=DEFAULT_REFINE_N_RANDOM, tol=DEFAULT_TOLERANCE, iterate=True
    ):

        if self.reduced_basis is None:
            self.reduced_basis = ReducedBasis().compute(
                self.model, Samples(self.model, n_random=n_random), tol=tol
            )
        self._validate_and_refine_basis(n_random, tol=tol, iterate=iterate)

        self.empirical_interpolant = EmpiricalInterpolant().compute(self.reduced_basis)

        return self

    def evaluate(self, params):
        _signal_at_nodes = self.model.compute(params, self.empirical_interpolant.nodes)
        return np.dot(_signal_at_nodes, np.real(self.empirical_interpolant.B_matrix))

    def write(self, filename):
        """Save ROM to file"""
        with h5py.File(filename, "w") as h5file:
            self.model.write(h5file)
            self.samples.write(h5file)
            if self.reduced_basis is not None:
                self.reduced_basis.write(h5file)
            if self.empirical_interpolant is not None:
                self.empirical_interpolant.write(h5file)

    @classmethod
    def from_file(cls, file_in):
        """Create a ROM instance from a file"""
        close_file = False
        if not isinstance(file_in, str):
            h5file = file_in
        else:
            h5file = h5py.File(file_in, "r")
            close_file = True
        model = RombusModel.from_file(h5file)
        samples = Samples.from_file(h5file)
        reduced_basis = ReducedBasis.from_file(h5file)
        empirical_interpolant = EmpiricalInterpolant.from_file(h5file)
        if close_file:
            h5file.close()
        return cls(
            model,
            samples,
            reduced_basis=reduced_basis,
            empirical_interpolant=empirical_interpolant,
        )

    def timing(self, samples):
        start_time = timeit.default_timer()
        for i, sample in enumerate(samples.samples):
            params_numpy = self.model.params.np2param(sample)
            _ = self.evaluate(params_numpy)
        return timeit.default_timer() - start_time

    def _validate_and_refine_basis(self, n_random, tol=DEFAULT_TOLERANCE, iterate=True):

        n_selected_greedy_points_global = np.iinfo(np.int32).max

        n_greedy_last = len(self.reduced_basis.greedypoints)
        n_greedy_last_global = mpi.COMM.allreduce(n_greedy_last, op=mpi4py.MPI.SUM)
        while True:
            # generate validation set by randomly sampling the parameter space
            new_samples = Samples(self.model, n_random=n_random)
            my_vs = self.model.generate_model_set(new_samples)

            # test validation set
            RB_transpose = np.transpose(self.reduced_basis.matrix)
            selected_greedy_points = []
            for i, validation_sample in enumerate(new_samples.samples):
                if self.model.model_dtype == complex:
                    proj_error = 1 - np.sum(
                        [
                            np.real(np.conjugate(d_i) * d_i)
                            for d_i in np.dot(my_vs[i], RB_transpose)
                        ]
                    )
                else:
                    proj_error = 1 - np.sum(np.dot(my_vs[i], RB_transpose) ** 2)
                if proj_error > tol:
                    selected_greedy_points.append(validation_sample)
            n_selected_greedy_points_global = mpi.COMM.allreduce(
                len(selected_greedy_points), op=mpi4py.MPI.SUM
            )
            if mpi.RANK_IS_MAIN:
                print(f"Number of samples added: {n_selected_greedy_points_global}")

            # add the inaccurate points to the original selected greedy
            # points and remake the basis
            self.samples.extend(selected_greedy_points)
            self.reduced_basis = ReducedBasis().compute(
                self.model, self.samples, tol=tol
            )
            n_greedy_new = len(self.reduced_basis.greedypoints)
            n_greedy_new_global = mpi.COMM.allreduce(n_greedy_new, op=mpi4py.MPI.SUM)

            if not iterate or n_greedy_new_global == n_greedy_last_global:
                break
            else:
                if mpi.RANK_IS_MAIN:
                    print(
                        f"Current number of accepted greedy points: {n_greedy_new_global}"
                    )
                n_greedy_last_global = n_greedy_new_global


class Samples(object):
    def __init__(self, model, filename=None, n_random=0):

        self.model = model

        # Needed for random number generation
        self.n_random = n_random
        self.random = None
        self.random_starting_state = None

        # Initialise samples
        self.n_samples = 0
        self.samples = []
        if filename:
            self._add_from_file(filename)
        if self.n_random > 0:
            self._add_random_samples(self.n_random)

    def extend(self, new_samples):

        self.samples.extend(new_samples)
        self.n_samples = self.n_samples + len(new_samples)

    def write(self, h5file):
        """Save samples to file"""

        h5_group = h5file.create_group("samples")
        self.model.write(h5_group)
        h5_group.create_dataset("samples", data=self.samples)
        h5_group.create_dataset("n_samples", data=self.n_samples)

    @classmethod
    def from_file(cls, file_in):
        """Create a ROM instance from a file"""

        close_file = False
        if not isinstance(file_in, str):
            h5file = file_in
        else:
            h5file = h5py.File(file_in, "r")
            close_file = True

        model_str = h5file["samples/model/model_str"].asstr()[()]
        model = RombusModel.load(model_str)
        samples = cls(model)
        samples.samples = [np.array(x) for x in h5file["samples/samples"]]
        samples.n_samples = np.int32(h5file["samples/n_samples"])
        if close_file:
            h5file.close()
        return samples

    def _add_from_file(self, filename_in: str):

        # dividing greedypoints into chunks
        if mpi.RANK_IS_MAIN:
            if filename_in.endswith(".npy"):
                samples = np.load(filename_in)
            elif filename_in.endswith(".csv"):
                samples = [
                    np.atleast_1d(x)
                    for x in np.genfromtxt(filename_in, delimiter=",", comments="#")
                ]
            else:
                raise Exception
        else:
            samples = None

        new_samples = self._decompose_samples(samples)
        n_new_samples = len(new_samples)

        self.samples.extend(new_samples)
        self.n_samples = self.n_samples + n_new_samples

    def _add_random_samples(self, n_samples):

        self.random = np.random.default_rng()
        self.random_starting_state = np.random.get_state()
        samples = []
        for _ in range(n_samples):
            # new_sample = np.ndarray(self.model.params.count, dtype=np.float64)
            # for i, param in enumerate(self.model.params):
            #    new_sample[i] = self.random.uniform(low=param.min, high=param.max)
            new_sample = self.model.params.generate_random_sample(self.random)
            samples.append(new_sample)

        new_samples = self._decompose_samples(samples)
        n_new_samples = len(new_samples)

        self.samples.extend(new_samples)
        self.n_samples = self.n_samples + n_new_samples

    def _decompose_samples(
        self,
        samples: List[np.ndarray],
    ):
        chunks = None
        if mpi.RANK_IS_MAIN:
            chunks = [[] for _ in range(mpi.SIZE)]
            for i, chunk in enumerate(samples):
                chunks[i % mpi.SIZE].append(chunk)

        return mpi.COMM.scatter(chunks, root=mpi.MAIN_RANK)


class ReducedBasis(object):
    def __init__(self, matrix=[], greedypoints=[], error_list=[]):
        """Make reduced basis

        FILENAME_IN is the 'greedy points' numpy file to take as input
        """

        self.matrix = matrix
        self.greedypoints = greedypoints
        self.error_list = error_list

    def compute(self, model, samples, tol=DEFAULT_TOLERANCE):

        if isinstance(model, str):
            self.model = RombusModel.load(model)
        elif isinstance(model, RombusModel):
            self.model = model
        else:
            raise Exception

        # Compute the model for each given sample
        my_ts = model.generate_model_set(samples)

        if mpi.RANK_IS_MAIN:
            print("Filling basis with greedy-algorithm")

        # hardcoding 1st model to be used to start the basis
        self._init_matrix(my_ts[0])
        basis_indicies = [0]

        # NOTE: NEED TO FIX; THE FIRST INDEX IS BEING ADDED TWICE HERE!
        error = np.inf
        iter = 1
        pc_matrix = []
        while error > tol:
            self.matrix, pc_matrix, error_data = self._add_next_model_to_basis(
                pc_matrix, my_ts, iter
            )
            err_rnk, err_idx, error = error_data

            # log and cache some data
            m = f">>> Iter {iter:003}: err {error:.1E} (rank {err_rnk:002}@idx{err_idx:003})"
            sys.stdout.write("\033[K" + m + "\r")

            basis_index = self._convert_to_basis_index(
                err_rnk, err_idx, samples.n_samples
            )
            self.error_list.append(error)
            basis_indicies.append(basis_index)

            # update iteration count
            iter += 1
        print("\n")

        self.matrix = np.asarray(self.matrix)
        self.greedypoints = [samples.samples[i] for i in basis_indicies]
        self.error_list = np.asarray(self.error_list)

        return self

        # if mpi.RANK_IS_MAIN:
        #    print("\nBasis generation complete!")
        #    if write_results:
        #        np.save("matrix", matrix)
        #        plot.errors(error_list)
        #        plot.basis(matrix)

    def write(self, h5file):
        """Save samples to file"""

        h5_group = h5file.create_group("reduced_basis")
        h5_group.create_dataset("matrix", data=self.matrix)
        h5_group.create_dataset("greedypoints", data=self.greedypoints)
        h5_group.create_dataset("error_list", data=self.error_list)

    @classmethod
    def from_file(cls, file_in):
        """Create a ROM instance from a file"""

        close_file = False
        if not isinstance(file_in, str):
            h5file = file_in
        else:
            h5file = h5py.File(file_in, "r")
            close_file = True

        matrix = np.array(h5file["reduced_basis/matrix"])
        greedypoints = [np.array(x) for x in h5file["reduced_basis/greedypoints"]]
        error_list = np.array(h5file["reduced_basis/error_list"])
        if close_file:
            h5file.close()
        return cls(matrix=matrix, greedypoints=greedypoints, error_list=error_list)

    def _add_next_model_to_basis(self, pc_matrix, my_ts, iter):
        # project training set on basis + get errors
        pc = misc.project_onto_basis(
            1.0, self.matrix, my_ts, iter - 1, self.model.model_dtype
        )
        pc_matrix.append(pc)

        projection_errors = list(
            1
            - np.einsum(
                "ij,ij->i", np.array(np.conjugate(pc_matrix)).T, np.array(pc_matrix).T
            )
        )

        # gather all errors (below is a list[ rank0_errors, rank1_errors...])
        all_rank_errors = mpi.COMM.gather(projection_errors, root=mpi.MAIN_RANK)

        # determine highest error
        if mpi.RANK_IS_MAIN:
            error_data = misc.get_highest_error(all_rank_errors)
            err_rank, err_idx, error = error_data
        else:
            error_data = None, None, None
        error_data = mpi.COMM.bcast(
            error_data, root=mpi.MAIN_RANK
        )  # share the error data with all nodes
        err_rank, err_idx, error = error_data

        # get model with the worst error
        worst_model = None
        if err_rank == mpi.MAIN_RANK:
            worst_model = my_ts[err_idx]  # no need to send
        elif mpi.RANK == err_rank:
            worst_model = my_ts[err_idx]
            mpi.COMM.send(worst_model, dest=mpi.MAIN_RANK)
        if worst_model is None and mpi.RANK_IS_MAIN:
            worst_model = mpi.COMM.recv(source=err_rank)

        # adding worst model to basis
        if mpi.RANK_IS_MAIN:
            # Gram-Schmidt to get the next basis and normalize
            self.matrix.append(misc.IMGS(self.matrix, worst_model, iter))

        # share the basis with ALL nodes
        matrix = mpi.COMM.bcast(self.matrix, root=mpi.MAIN_RANK)
        return matrix, pc_matrix, error_data

    def _convert_to_basis_index(self, rank_number, rank_idx, rank_count):
        ranks_till_err_rank = [i for i in range(rank_number)]
        idx_till_err_rank = np.sum([rank_count[i] for i in ranks_till_err_rank])
        return int(rank_idx + idx_till_err_rank)

    def _init_matrix(self, init_model):
        # init the baisis (matrix) with 1 model from the training set to start
        if mpi.RANK_IS_MAIN:
            self.matrix = [init_model]
        else:
            self.matrix = None
        # share the basis with ALL nodes
        self.matrix = mpi.COMM.bcast(self.matrix, root=mpi.MAIN_RANK)


class EmpiricalInterpolant(object):
    def __init__(self, B_matrix=None, nodes=None):
        """Initialise empirical interpolant"""

        self.B_matrix = B_matrix
        self.nodes = nodes

    def compute(self, reduced_basis):
        """Compute empirical interpolant"""

        # RB = RB[0 : len(RB)]
        if mpi.RANK_IS_MAIN:
            print("Computing empirical interpolant")
        eim = algorithms.StandardEIM(
            reduced_basis.matrix.shape[0], reduced_basis.matrix.shape[1]
        )
        eim.make(reduced_basis.matrix)
        domain = reduced_basis.model.domain
        self.nodes = domain[eim.indices]
        self.nodes, self.B_matrix = zip(*sorted(zip(self.nodes, eim.B)))

        return self

    def write(self, h5file):
        """Save empirical interpolant to file"""

        h5_group = h5file.create_group("empirical_interpolant")
        h5_group.create_dataset("B_matrix", data=self.B_matrix)
        h5_group.create_dataset("nodes", data=self.nodes)

    @classmethod
    def from_file(cls, file_in):
        """Create a ROM instance from a file"""

        close_file = False
        if not isinstance(file_in, str):
            h5file = file_in
        else:
            h5file = h5py.File(file_in, "r")
            close_file = True
        B_matrix = np.array(h5file["empirical_interpolant/B_matrix"])
        nodes = np.array(h5file["empirical_interpolant/nodes"])
        if close_file:
            h5file.close()
        return cls(B_matrix=B_matrix, nodes=nodes)



import sys

import click

import collections

import rombus.core as core
import rombus.plot as plot

from rombus.model import RombusModel

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])
FLEX_CONTEXT_SETTINGS = dict(
    help_option_names=["-h", "--help"],
    ignore_unknown_options=True,
    allow_extra_args=True,  # needed for the passing of model parameters
)


class _OrderedGroup(click.Group):
    """This class is used for to ensure that the ordering of the CLI subcommands
    are in code-order in the CLI help and documentation."""

    def __init__(self, name=None, commands=None, **attrs):
        super(_OrderedGroup, self).__init__(name, commands, **attrs)
        #: the registered subcommands by their exported names.
        self.commands = commands or collections.OrderedDict()

    def list_commands(self, ctx):
        return self.commands


@click.group(cls=_OrderedGroup, context_settings=CONTEXT_SETTINGS)
@click.pass_context
def cli(ctx):
    """Perform greedy algorythm operations with Rombus"""

    # ensure that ctx.obj exists and is a dict (in case `cli()` is called
    # by means other than the `if` block below)
    ctx.ensure_object(dict)


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.argument("project_name", type=str)
def quickstart(project_name):
    """Write a project template to build a new project from"""

    RombusModel.write_project_template(project_name)


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.argument("model", type=str)
@click.argument("filename_samples", type=click.Path(exists=True))
@click.option(
    "--out",
    "-o",
    default="MODEL_BASENAME.hdf5",
    type=click.Path(exists=False),
    show_default=True,
    help="Output HDF5 filename",
)
@click.option(
    "--do_step",
    type=click.Choice(["RB", "EI"], case_sensitive=False),
    default=None,
    help="Do only one step: RB=reduced basis; EI=empirical interpolant",
)
@click.pass_context
def build(ctx, model, filename_samples, out, do_step):
    """Build a reduced order model

    FILENAME_IN is the 'greedy points' numpy or csv file to take as input
    """

    # Load model
    model = RombusModel.load(model)

    # Load samples
    samples = core.Samples(model, filename=filename_samples)

    # Build ROM
    ROM = core.ROM(model, samples).build(do_step=do_step)

    # Write ROM
    if out == "MODEL_BASENAME.hdf5":
        filename_out = f"{model.model_basename}.hdf5"
    else:
        filename_out = out
    ROM.write(filename_out)


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.argument("filename_ROM", type=click.Path(exists=True))
@click.pass_context
def refine(ctx, filename_rom):
    """Refine parameter sampling to impove a reduced order model"""

    # Build model and refine it
    ROM = core.ROM.from_file(filename_rom).refine()

    # Write results
    filename_split = filename_rom.rsplit(".", 1)
    filename_out = f"{filename_split[0]}_refined.{filename_split[1]}"
    ROM.write(filename_out)


@cli.command(context_settings=FLEX_CONTEXT_SETTINGS)
@click.argument("filename_ROM", type=str)
@click.argument("parameters", nargs=-1, type=click.UNPROCESSED)
@click.pass_context
def evaluate(ctx, filename_rom, parameters):
    """Evaluate a reduced order model and compare it to truth

    PARAMETERS is a list of parameter values of the form A=VAL B=VAL ..."""

    # Read ROM
    ROM = core.ROM.from_file(filename_rom)

    # Parse the model parameters, which should have been given as arguments
    model_params = ROM.model.parse_cli_params(parameters)

    # Generate plot
    plot.compare_rom_to_true(ROM, model_params)


@cli.command(context_settings=FLEX_CONTEXT_SETTINGS)
@click.argument("filename_ROM", type=str)
@click.option(
    "-n",
    "--n_samples",
    type=int,
    default=100,
    help="Number of samples to use for timing",
)
@click.pass_context
def timing(ctx, filename_rom, n_samples):
    """Compute timing information for a ROM and it's source model"""

    # Read ROM
    ROM = core.ROM.from_file(filename_rom)

    # Generate the samples to be used
    timing_sample = core.Samples(ROM.model, n_random=n_samples)

    # Generate timing information for model
    timing_model = ROM.model.timing(timing_sample)

    # Generate timing information for ROM
    timing_ROM = ROM.timing(timing_sample)

    # Report results
    print(
        f"Timing information for ROM:   {timing_ROM:.2e}s for {n_samples} calls ({timing_ROM/n_samples:.2e} per sample)."
    )
    print(
        f"Timing information for model: {timing_model:.2e}s for {n_samples} calls ({timing_model/n_samples:.2e} per sample)."
    )
    if timing_ROM > timing_model:
        print(f"ROM is {timing_ROM/timing_model:.2f}X slower than the source model.")
    else:
        print(f"ROM is {timing_model/timing_ROM:.2f}X faster than the source model.")


if __name__ == "__main__":
    cli(obj={})
    sys.exit()



# --- training.py ---

"""
    Classes for building training sets
"""

__author__ = "Chad Galley <crgalley@gmail.com>"

import random

import numpy as np
from numpy import asarray


##############################################
class TrainingTools:
    def __init__(self):
        pass

    def equal_lens(self, *arrs):
        lens = map(len, arrs)
        if len(set(lens)) > 1:
            return False
        else:
            return True

    def check_lens(self, *arrs):
        if not self.equal_lens(*arrs):
            raise Exception("Unequal array lengths")
        pass

    def make_arrs(self, pmin, pmax, Nmu):
        """Make list of 1d arrays for each given parameter range"""
        if np.shape(Nmu):
            return [np.linspace(pmin[jj], pmax[jj], Nmu[jj]) for jj in range(len(pmin))]
        else:
            return [np.linspace(pmin[jj], pmax[jj], Nmu) for jj in range(len(pmin))]

    def meshgrid2(self, *arrs):
        """Multi-dimensional version of numpy's meshgrid"""
        arrs = tuple(reversed(arrs))
        lens = map(len, arrs)
        dim = len(arrs)
        sz = 1
        for s in lens:
            sz *= s

        ans = []
        for i, arr in enumerate(arrs):
            slc = [1] * dim
            slc[i] = lens[i]
            arr2 = asarray(arr).reshape(slc)
            for j, sz in enumerate(lens):
                if j != i:
                    arr2 = arr2.repeat(sz, axis=j)
            ans.append(arr2)

        return ans[::-1]

    def meshgrid3(self, *arrs):
        """Multi-dimensional version of numpy's meshgrid,
        excluding symmetric tuples from the square grid"""
        arrs = tuple(reversed(arrs))
        lens = map(len, arrs)
        inds = [np.triu_indices(ll) for ll in lens]
        dim = len(arrs)
        sz = 1
        for s in lens:
            sz *= s

        ans = []
        for i, arr in enumerate(arrs):
            slc = [1] * dim
            slc[i] = lens[i]
            arr2 = asarray(arr).reshape(slc)
            for j, sz in enumerate(lens):
                if j != i:
                    arr2 = arr2.repeat(sz, axis=j)
            ans.append(arr2[inds[i]])

        return ans[::-1]

    def tuple_to_vstack(self, arr):
        return np.vstack(map(np.ravel, tuple(arr)))

    def rescale(self, x):
        """Scale [a,b] to [0,1]"""
        a = x[0]
        b = x[-1]
        return (x - a) / (b - a)

    def invrescale(self, x, a, b):
        """Scale [0,1] to [a,b]"""
        if float(x[0]) != 0.0 or float(x[-1]) != 1.0:
            raise Exception("Expected a [0,...,1] array")
        return a + (b - a) * x


##############################################
class Uniform(TrainingTools):
    def __init__(self):
        TrainingTools.__init__(self)
        self.switch_dict = {
            "grid": self.grid,
            "symgrid": self.sym_grid,
            "1d": self.oned,
            "1d-constrained": self.oned_constrained,
        }
        self.options = self.switch_dict.keys()

    def __call__(self, params, training_type):
        return self.switch_dict[training_type](*params)

    def oned(self, pmin, pmax, Nmu):
        """Uniformly spaced 1d array"""
        return np.linspace(pmin, pmax, num=Nmu)

    def oned_constrained(self, pmin, pmax, Nmu):
        """Uniformly spaced 1d array satisfying a constraint"""
        arrs = self.make_arrs(pmin, pmax, Nmu)
        return self.tuple_to_vstack(arrs)

    def grid(self, pmin, pmax, Nmu):
        """Uniformly spaced n-dimensional array"""
        arrs = self.make_arrs(pmin, pmax, Nmu)
        return self.tuple_to_vstack(self.meshgrid2(*arrs))

    def sym_grid(self, pmin, pmax, Nmu):
        """Uniformly spaced n-dimensional array with
        symmetric tuples removed from square grid
        """
        arrs = self.make_arrs(pmin, pmax, Nmu)
        if np.shape(Nmu):
            if len(set(Nmu)) > 1:
                raise Exception(
                    "Number of points in each parameter dimension are not equal."
                )
        return self.tuple_to_vstack(self.meshgrid3(*arrs))


##############################################
class Random(TrainingTools):
    def __init__(self):
        TrainingTools.__init__(self)
        self.switch_dict = {"uniform": self.uniform, "nonuniform": self.nonuniform}
        self.options = self.switch_dict.keys()

    def __call__(self, params, training_type):
        return self.switch_dict[training_type](*params)

    def uniform(self, pmin, pmax, Nmu):
        if np.shape(pmin):
            lens = map(len, [pmin, pmax])
            ans = []
            if self.equal_lens(pmin, pmax):
                for ii in range(lens[0]):
                    arr = [random.uniform(pmin[ii], pmax[ii]) for jj in range(Nmu)]
                    ans.append(arr)
            return self.tuple_to_vstack(ans)
        else:
            return self.tuple_to_vstack(
                [[random.uniform(pmin, pmax)] for jj in range(Nmu)]
            )

    def nonuniform(self):
        pass


##############################################
class NonUniform(TrainingTools):
    def __init__(self):
        TrainingTools.__init__(self)
        self.switch_dict = {"1d": self.oned, "1d-constrained": self.oned_constrained}
        self.options = self.switch_dict.keys()

    def __call__(self, fn, params, training_type):
        return self.switch_dict[training_type](fn, *params)

    def oned(self, fn, pmin, pmax, Nmu):
        """Nonuniform 1d training space made from an input function"""
        x = np.linspace(pmin, pmax, num=Nmu)
        u = fn(self.rescale(x))
        return self.invrescale(u, pmin, pmax)

    def oned_constrained(self, fn, pmin, pmax, Nmu):
        """Nonuniform 1d, constrained training space made from an input function"""
        mu = np.transpose([pmin, pmax, Nmu])
        arrs = []
        for mm in mu:
            arrs.append(self.oned(fn, *mm))
        return self.tuple_to_vstack(arrs)



# --- algorithms.py ---

"""
Standard algorithms for::
   -- generating an orthonormal basis from a set of functions
   -- building a reduced spline for 1d functions (removed)
   -- building a reduced basis
   -- empirical interpolation
   -- reduced-order quadrature rules (future)
"""

__author__ = "Chad Galley <crgalley@gmail.com>"

import time

import numpy as np

from . import eim
from . import greedy
from . import lib


##############################################
class Basis(greedy.IteratedModifiedGramSchmidt):
    def __init__(self, hs, inner, inner_type, normsQ=False, dtype="complex"):
        greedy.IteratedModifiedGramSchmidt.__init__(self, inner, inner_type)

        self.Nbasis, self.Nquads = np.shape(hs)
        self.functions = hs

        self.normsQ = normsQ
        if self.normsQ:
            self.norms = lib.malloc(dtype, self.Nbasis)

        self.basis = lib.malloc(dtype, self.Nbasis, self.Nquads)
        pass

    def iter(self, step, h, a=0.5, max_iter=3):
        ans = self.add_basis(h, self.basis[:step], a=a, max_iter=max_iter)

        if self.normsQ:
            self.basis[step + 1], self.norms[step + 1] = ans
        else:
            self.basis[step + 1], _ = ans

        pass

    def make(self, a=0.5, max_iter=3, timerQ=False):
        """Find the corresponding orthonormal set of basis functions."""

        self.basis[0] = self.inner.normalize(self.functions[0], self.inner_type)
        if self.normsQ:
            self.norms[0] = self.inner.norm(self.functions[0], self.inner_type)

        if timerQ:
            t0 = time.time()

        for ii in range(1, self.Nbasis):
            self.iter(ii, self.functions[ii], a=a, max_iter=max_iter)

        if timerQ:
            print("\nElapsed time =", time.time() - t0)

        if self.normsQ:
            return [np.array(self.basis), self.norms]
        else:
            return np.array(self.basis)
        pass


##############################################
# class MakeRS(ReducedSpline1d, Spline1d):
#
#     def __init__(self, x, deg=3, dtype='double'):
#         ReducedSpline1d.__init__(self)
#         self.x = np.sort(x)
#         self.deg = deg
#         self.nodes = np.zeros(len(x), dtype='double')
#         self.values = np.zeros(len(x), dtype=dtype)
#         pass
#
#     def seed(self, x0, fn, args=None):
#         self.xmax = self.x.max()
#         self.nodes[0] = x0
#         self.values[0] = lib.fneval(x0, fn, args=args)
#         pass
#
#     def iter(self, step, delta, x, fn, args=None):
#         if x+delta < self.xmax:
#             self.nodes[step+1] = x+delta
#             self.values[step+1] = lib.fneval(x+delta, fn, args=args)
#             return x+delta, step+1
#         else:
#             self.nodes[step+1] = self.xmax
#             self.values[step+1] = lib.fneval(self.xmax, fn, args=args)
#             return self.xmax, step+1
#         pass
#
#     def make(self, tol, fn, args=None):
#
#         self.seed(self.x.min(), fn, args=args)
#
#         #xsorted = np.sort(self.x)
#         ctr = 0
#         xx = self.nodes[0]
#         while xx <= self.xmax:
#
#             deriv_n = self.autodiff(xx, fn, n=self.deg+1, args=args)
#             delta = self.delta(tol, deriv_n, self.deg+1)
#
#             xx, ctr = self.iter(ctr, delta, xx, fn, args=args)
#
#             if xx+delta > self.xmax:
#                 break
#
#         self.trim(ctr+1)
#         self.size = len(self.nodes)
#
#         # Make the spline interpolant
#         spline = Spline1d(self.nodes, self.values, k=self.deg)
#         self.fit = spline.fit
#         self.fitparams = spline.fitparams
#         pass
#
#     def trim(self, num):
#         """Trim zeros from remaining entries"""
#         self.nodes = self.nodes[:num+1]
#         self.values = self.values[:num+1]
#         pass


##############################################
class StandardRB(greedy.ReducedBasis, greedy.IteratedModifiedGramSchmidt):
    """Class for standard reduced basis greedy algorithm.

    Input
    -----
       Nbasis       -- number of basis functions
       Npoints       -- number of training set points
       Nquads       -- number of quadrature nodes used for inner products
       inner       -- method of InnerProduct instance
       inner_type -- type of inner product (e.g., 'complex')
       norms       -- array of training set function norms (default None)
       dtype       -- data type of functions (default 'complex')

    Functions
    ---------
       seed -- seed the greedy algorithm
       iter -- one iteration of the greedy algorithm
       make -- implement the greedy algorithm from beginning to end
       trim -- trim zeros from remaining allocated entries

    Examples
    --------
       Create an instance of the StandardRB class for functions with
       unit norm::

         >>> rb = rp.StandardRB(Nbasis, Npoints, Nquads, inner, inner_type)

       Let T be the training space of functions, 0 be the seed index,
       and 1e-12 be the tolerance. The standard reduced basis greedy
       algorithm for normalized functions is::

         >>> rb.seed(0, T)
         >>> for i in range(Nbasis):
         >>> ...if rb.errors[i] <= 1e-12:
         >>> ......break
         >>> ...rb.iter(i,T)
         >>> rb.trim(i)

       For convenience, this algorithm is equivalently implemented in
       `make`::

         >>> rb.make(0, T, 1e-12)

       Let T' be a different training space. The greedy algorithm can
       be run again on T' using::

         >>> rb.make(0, T', 1e-12)

       or, alternatively, at each iteration using::

         >>> ...rb.iter(i,T')

       in the for-loop above.

    """

    def __init__(
        self,
        Nbasis,
        Npoints,
        Nquads,
        inner,
        inner_type,
        Nmodes=1,
        norms=None,
        dtype="complex",
    ):
        greedy.ReducedBasis.__init__(self, inner, inner_type)
        greedy.IteratedModifiedGramSchmidt.__init__(self, inner, inner_type)

        if norms is None:
            self.norms = np.ones(Npoints, dtype="double")
        else:
            self.norms = norms
        self.ctr = 1

        # Allocate memory for numpy arrays
        self.malloc(Nbasis, Npoints, Nquads, Nmodes=Nmodes, dtype=dtype)
        pass

    def seed(self, index_seed, trsp):
        """Seed the greedy algorithm.

        Seeds the first entries in the errors, indices, basis, and alpha arrays
        for use with the standard greedy algorithm for producing a reduced basis
        representation.

        Input
        -----
        index_seed     -- array index for seed point in training set
        trsp         -- the training space of functions

        Examples
        --------

        If rb is an instance of StandardRB, 0 is the array index associated
        with the seed, and T is the training set then do::

          >>> rb.seed(0, T)

        """

        self.errors[0] = np.max(self.norms) ** 2
        self.indices[0] = index_seed
        self.basis[0] = trsp[index_seed] / self.norms[index_seed]
        self.basisnorms[0] = self.norms[index_seed]
        self.alpha[0] = self.alpha_arr(self.basis[0], trsp)
        pass

    def iter(self, step, errs, trsp):
        """One iteration of standard reduced basis greedy algorithm.

        Updates the next entries of the errors, indices, basis, and
        alpha arrays.

        Input
        -----
        step -- current iteration step
        trsp -- the training space of functions

        Examples
        --------

        If rb is an instance of StandardRB and iter=13 is the 13th
        iteration of the greedy algorithm then the following code
        snippet generates the next (i.e., 14th) entry of the errors,
        indices, basis, and alpha arrays::

          >>> rb.iter(13)

        """

        self.errors[step + 1] = np.max(errs)
        self.indices[step + 1] = np.argmax(errs)
        self.basis[step + 1], self.basisnorms[step + 1] = self.add_basis(
            trsp[self.indices[step + 1]], self.basis[: step + 1]
        )
        self.alpha[step + 1] = self.alpha_arr(self.basis[step + 1], trsp)
        pass

    def make(self, index_seed, trsp, tol, verbose=True, timerQ=False):
        """Make a reduced basis using the standard greedy algorithm.

        Input
        -----
        index_seed     -- array index for seed point in training set
        trsp         -- the training space of functions
        tol            -- tolerance that terminates the greedy algorithm
        verbose        -- print projection errors to screen (default True)
        timer        -- print elapsed time (default False)

        Examples
        --------
        If rb is the StandardRB class instance, 0 the seed index, and
        T the training set then do::

          >>> rb.make(0, T, 1e-12)

        To prevent displaying any print to screen, set the `verbose`
        keyword argument to `False`::

          >>> rb.make(0, T, 1e-12, verbose=False)

        """

        # Seed the greedy algorithm
        self.seed(index_seed, trsp)
        flag = 0

        # The standard greedy algorithm with fixed training set
        if verbose:
            print("\nIter", "\t", "Error")
        if timerQ:
            t0 = time.time()
        for ctr in range(len(self.errors) - 1):
            if verbose:
                print(ctr + 1, "\t", self.errors[ctr] / self.errors[0])

            # Check if tolerance is met
            if self.errors[ctr] <= tol:
                flag = 1
                break

            # Single iteration and update errors, indices, basis, alpha arrays
            errs = self.proj_errors_from_alpha(self.alpha[: ctr + 1], norms=self.norms)
            self.iter(ctr, errs, trsp)

        if timerQ:
            print("\nElapsed time =", time.time() - t0)

        # Actual number of basis functions
        if flag == 1:
            self.size = ctr
        else:
            self.size = ctr + 2

        # Trim excess allocated entries
        self.trim(self.size)
        pass

    def projection(self, h):
        return self.rb_project(h, self.basis)

    def trim(self, num):
        """Trim zeros from remaining entries"""
        self.errors = self.errors[:num]
        self.indices = self.indices[:num]
        self.basis = self.basis[:num]
        self.alpha = self.alpha[:num]
        pass


##############################################
class StandardEIM(eim.EmpiricalInterpolation):
    def __init__(self, Nbasis, Nquads, Nmodes=1, dtype="complex"):
        eim.EmpiricalInterpolation.__init__(self)

        # Allocate memory for numpy arrays
        self.malloc(Nbasis, Nquads, Nmodes=Nmodes, dtype=dtype)
        self.modes = Nmodes
        self.quads = Nquads

    def seed(self, e):
        """Seed the algorithm"""
        self.indices[0] = np.argmax(np.abs(e))
        self.R[0] = e
        self.invV[:1, :1] = self.next_invV_col(self.R[:1], self.indices[:1])
        pass

    def iter(self, step, e):
        """One iteration in the empirical interpolation greedy algorithm"""

        ctr = step + 1

        # Compute interpolant residual
        c = self.coefficient(self.invV[:ctr, :ctr], e, self.indices[:ctr])
        r = self.residual(e, c, self.R[:ctr])

        # Update
        self.indices[ctr] = np.argmax(np.abs(r))
        self.R[ctr] = r
        self.invV[:, ctr][: ctr + 1] = self.next_invV_col(
            self.R[: ctr + 1], self.indices[: ctr + 1]
        )
        pass

    def make(self, basis, verbose=True, timerQ=False):
        """Make an empirical interpolant using the standard greedy algorithm"""

        # Seed the greedy algorithm
        self.seed(basis[0])

        # EIM algorithm with reduced complexity for inverting the van der Monde matrix
        if verbose:
            print("\nIter", "\t", "Indices")
        if timerQ:
            t0 = time.time()
        dim = len(basis[:, 0])
        for ctr in range(dim - 1):
            if verbose:
                print(ctr + 1, "\t", self.indices[ctr])

            # Single iteration
            self.iter(ctr, basis[ctr + 1])

        if timerQ:
            print("\nElapsed time =", time.time() - t0)

        # Compute interpolant matrix 'B'
        # if flag == 1:
        #     self.size = ctr
        # else:
        #     self.size = ctr+2
        # self.trim(ctr+1)
        self.trim(ctr + 2)
        self.make_interpolant()
        self.size = len(self.indices)
        pass

    def make_interpolant(self):
        self.B = self.eim_interpolant(self.invV, self.R)
        pass

    def interpolate(self, h):
        return self.eim_interpolate(h, self.indices, self.B)

    def trim(self, num):
        """Trim zeros from remaining entries"""
        if num > 0:
            self.indices = self.indices[:num]
            self.R = self.R[:num]
            self.invV = self.invV[:num, :num]
            self.B = self.B[:num, :]
        pass


##############################################
class StandardRBEIM(greedy.ReducedBasis, eim.EmpiricalInterpolation):
    def __init__(
        self,
        Nbasis,
        Npoints,
        Nquads,
        inner,
        inner_type,
        Nmodes=1,
        norms=None,
        dtype="complex",
    ):
        self.rb = StandardRB(
            Nbasis,
            Npoints,
            Nquads,
            inner,
            inner_type,
            Nmodes=Nmodes,
            norms=norms,
            dtype=dtype,
        )
        self.eim = StandardEIM(Nbasis, Nquads, dtype=dtype)
        greedy.ReducedBasis.__init__(self, inner, inner_type)
        eim.EmpiricalInterpolation.__init__(self)

    def seed(self, rb_index_seed, trsp):
        """Seed the algorithms"""
        self.rb.seed(rb_index_seed, trsp)
        self.eim.seed(self.rb.basis[0])
        pass

    def iter(self, step, trsp, e):
        """One iteration in the reduced basis and empirical
        interpolation greedy algorithms
        """
        self.rb.iter(step, trsp)
        self.eim.iter(step, e)

    def make(self, rb_index_seed, trsp, tol, verbose=True):
        """Make a reduced basis and an empirical interpolant simultaneously"""

        # Seed the greedy algorithm
        self.seed(rb_index_seed, trsp)

        # Greedy algorithm
        if verbose:
            print("\nIter", "\t", "Error", "\t", "Indices")
        dim = len(self.rb.basis[:, 0])
        for ctr in range(dim - 1):
            if verbose:
                print(ctr + 1, "\t", self.rb.errors[ctr], "\t", self.eim.indices[ctr])

            # Check if tolerance is met
            if self.rb.errors[ctr] <= tol:
                break

            # Single iteration
            self.iter(ctr, trsp, self.rb.basis[ctr + 1])

        # Compute interpolant matrix 'B'
        self.rb.trim(ctr)
        self.eim.trim(ctr)
        self.eim.make_interpolant()
        self.size = ctr
        pass



# --- eim.py ---

"""
    Classes for the empirical interpolation method
"""

__author__ = "Chad Galley <crgalley@gmail.com>"
import numpy as np

# from __init__ import np
from scipy.linalg.lapack import get_lapack_funcs

# from scipy.linalg import solve_triangular
from . import lib


##############################################
class LinAlg:
    """Linear algebra functions needed for empirical interpolation class"""

    def __init__(self):
        pass

    # This is scipy.linalg's source code. For some reason my scipy doesn't recognize
    # the check_finite option, which may help with speeding up. Because this may be an
    # issue related to the latest scipy.linalg version, I'm reproducing that code here
    # to guarantee future compatibility.
    def solve_triangular(
        self,
        a,
        b,
        trans=0,
        lower=False,
        unit_diagonal=False,
        overwrite_b=False,
        debug=False,
        check_finite=True,
    ):
        """
        Solve the equation `a x = b` for `x`, assuming a is a triangular matrix.

        Parameters
        ----------
        a : (M, M) array_like
            A triangular matrix
        b : (M,) or (M, N) array_like
            Right-hand side matrix in `a x = b`
        lower : boolean
            Use only data contained in the lower triangle of `a`.
            Default is to use upper triangle.
        trans : {0, 1, 2, 'N', 'T', 'C'}, optional
            Type of system to solve:

            ========  =========
            trans     system
            ========  =========
            0 or 'N'  a x  = b
            1 or 'T'  a^T x = b
            2 or 'C'  a^H x = b
            ========  =========
        unit_diagonal : bool, optional
            If True, diagonal elements of `a` are assumed to be 1 and
            will not be referenced.
        overwrite_b : bool, optional
            Allow overwriting data in `b` (may enhance performance)
        check_finite : bool, optional
            Whether to check that the input matrices contain only finite numbers.
            Disabling may give a performance gain, but may result in problems
            (crashes, non-termination) if the inputs do contain infinities or NaNs.

        Returns
        -------
        x : (M,) or (M, N) ndarray
            Solution to the system `a x = b`.  Shape of return matches `b`.

        Raises
        ------
        Exception
            If `a` is singular

        Notes
        -----
        .. versionadded:: 0.9.0

        """

        if check_finite:
            a1, b1 = map(np.asarray_chkfinite, (a, b))
        else:
            a1, b1 = map(np.asarray, (a, b))
        if len(a1.shape) != 2 or a1.shape[0] != a1.shape[1]:
            raise ValueError("expected square matrix")
        if a1.shape[0] != b1.shape[0]:
            raise ValueError("incompatible dimensions")
        overwrite_b = False  # overwrite_b or _datacopied(b1, b)
        if debug:
            print("solve:overwrite_b=", overwrite_b)
        trans = {"N": 0, "T": 1, "C": 2}.get(trans, trans)
        (trtrs,) = get_lapack_funcs(("trtrs",), (a1, b1))
        x, info = trtrs(
            a1,
            b1,
            overwrite_b=overwrite_b,
            lower=lower,
            trans=trans,
            unitdiag=unit_diagonal,
        )

        if info == 0:
            return x
        if info > 0:
            raise Exception(
                "singular matrix: resolution failed at diagonal %s" % (info - 1)
            )
        raise ValueError("illegal value in %d-th argument of internal trtrs" % -info)

    def transpose(self, a):
        dim = a.shape
        if len(dim) != 2:
            raise ValueError("Expected a matrix")
        aT = np.zeros(dim[::-1], dtype=a.dtype)
        for ii, aa in enumerate(a):
            aT[:, ii] = a[ii]
        return aT


##############################################
class EmpiricalInterpolation(LinAlg):
    """
    Class for building an empirical interpolant
    """

    def __init__(self):
        LinAlg.__init__(self)

    def malloc(self, Nbasis, Nquads, Nmodes=1, dtype="complex"):
        self.indices = lib.malloc("int", Nbasis)
        self.invV = lib.malloc(dtype, Nbasis, Nbasis)
        self.R = lib.malloc(dtype, Nbasis, Nquads)
        self.B = lib.malloc(dtype, Nbasis, Nquads)
        pass

    def coefficient(self, invV, e, indices):
        return np.dot(invV.T, e[indices])

    def residual(self, e, c, R):
        """Difference between a basis function 'e' and its empirical interpolation"""
        return e - np.dot(c, R)

    def next_invV_col(self, R, indices, check_finite=False):
        b = np.zeros(len(indices), dtype=R.dtype)
        b[-1] = 1.0
        return self.solve_triangular(
            R[:, indices], b, lower=False, check_finite=check_finite
        )
        # return solve_triangular(
        #    R[:,indices], b, lower=False, check_finite=check_finite)

    def eim_interpolant(self, invV, R):
        """The empirical interpolation matrix 'B'"""
        return np.dot(invV, R)

    def eim_interpolate(self, h, indices, B):
        """Empirically interpolate a function"""
        dim = np.shape(h)
        if len(dim) == 1:
            return np.dot(h[indices], B)
        elif len(dim) > 1:
            return np.array([np.dot(h[ii][indices], B) for ii in range(dim[0])])



# -- lib.py --

"""Basic functions used by several subpackages."""

__author__ = "Chad Galley <crgalley@gmail.com>"

import numpy as np
from scipy.special import factorial


def malloc(dtype, *nums):
    """Allocate some memory with given dtype"""
    return np.zeros(tuple(nums), dtype=dtype)


def malloc_more(arr, num_more):
    """Allocate more memory to append to arr"""
    dim = len(arr.shape)
    if dim == 1:
        return np.hstack([arr, malloc(arr.dtype, num_more)])
    elif dim == 2:
        # Add num_extra rows to arr
        shape = arr.shape
        return np.vstack([arr, malloc(arr.dtype, num_more, shape[1])])
    else:
        raise Exception("Expected a vector or matrix.")


def trim(arr, num):
    return arr[:num]


def scale_ab_to_cd(x, c, d):
    """Scale [a,b] to [c,d]"""
    a = x[0]
    b = x[-1]
    a, b, c, d = map(float, [a, b, c, d])
    return (d - c) / (b - a) * x - (a * d - b * c) / (b - a)


def scale_ab_to_01(x):
    """Scale [a,b] to [0,1]"""
    interval = scale_ab_to_cd(x, 0, 1)
    return np.abs(interval)


def scale_01_to_ab(x, a, b):
    """Scale [0,1] to [a,b]"""
    if np.allclose(float(x[0]), 0.0) and np.allclose(float(x[-1]), 1.0):
        return scale_ab_to_cd(np.abs(x), a, b)
    else:
        raise Exception("Expected a [0,...,1] array")


def fneval(x, fn, args=None):
    if args is None:
        return fn(x)
    elif len(args) >= 1:
        return fn(x, *args)
    else:
        raise Exception("Function arguments not a 1d array.")
    pass


def get_arg(a, a0):
    """Get argument at which a0 occurs in array a"""
    return abs(a - a0).argmin()


def map_intervals(x, a, b):
    """Map array x to interval [a,b]"""
    M = (b - a) / (np.max(x) - np.min(x))
    B = a - M * np.min(x)
    return M * x + B


def choose(top, bottom):
    """Combinatorial choose function"""
    return factorial(top) / factorial(bottom) / factorial(top - bottom)


class Empty:
    """A class with no purpose other than for attaching class attributes"""

    def __init__(self):
        pass



from rombus._core import misc


def test_highest_error():
    err_list = [[0, 0, 1], [0, 2], [100, 0]]
    rank, idx, err = misc.get_highest_error(err_list)
    assert rank == 2
    assert idx == 0
    assert err == 100



import pytest


def pytest_collection_modifyitems(config, items):
    keywordexpr = config.option.keyword
    markexpr = config.option.markexpr
    if keywordexpr or markexpr:
        return  # let pytest handle this

    skip_lalsuite = pytest.mark.skip(reason="lalsuite not selected")
    for item in items:
        if "lalsuite" in item.keywords:
            item.add_marker(skip_lalsuite)



# --- greedy.py ---

"""
    Classes for building reduced basis greedy algorithms
"""

__author__ = "Chad Galley <crgalley@gmail.com>"

import numpy as np

# from __init__ import np
# import adipy
from . import lib


##############################################
class IteratedModifiedGramSchmidt:
    """Iterated modified Gram-Schmidt algorithm for building an orthonormal basis.
    Algorithm from Hoffman, `Iterative Algorithms for Gram-Schmidt Orthogonalization`.
    """

    def __init__(self, inner, inner_type):
        self.inner = inner
        self.inner_type = inner_type

    def add_basis(self, h, basis, a=0.5, max_iter=3):
        """Given a function, h, find the corresponding basis
        function orthonormal to all previous ones
        """
        norm = self.inner.norm(h, self.inner_type)
        e = h / norm
        flag, ctr = 0, 1
        while flag == 0:
            for b in basis:
                e -= b * self.inner.dot(b, e, self.inner_type)
            new_norm = self.inner.norm(e, self.inner_type)

            # Iterate, if necessary
            if new_norm / norm <= a:
                norm = new_norm
                ctr += 1
                if ctr > max_iter:
                    print(
                        (
                            f"Max number of iterations ({str(max_iter)}) "
                            "reached in iterated Gram-Schmidt. Basis "
                            "may not be orthonormal."
                        )
                    )
            else:
                flag = 1

        return [e / new_norm, new_norm]

    def make_basis(self, hs, norms=False, a=0.5, max_iter=3):
        """Given a set of functions, hs, find the corresponding
        orthonormal set of basis functions.
        """

        dim = np.shape(hs)
        basis = np.zeros(dim, dtype=hs.dtype)
        basis[0] = self.inner.normalize(hs[0], self.inner_type)
        if norms:
            norm = np.zeros(dim[0], dtype="double")
            norm[0] = self.inner.norm(hs[0], self.inner_type)

        for ii in range(1, dim[0]):
            if norms:
                basis[ii], norm[ii] = self.add_basis(
                    hs[ii], basis[:ii], a=a, max_iter=max_iter
                )
            else:
                basis[ii], _ = self.add_basis(
                    hs[ii], basis[:ii], a=a, max_iter=max_iter
                )

        if norms:
            return [np.array(basis), norm]
        else:
            return np.array(basis)


##############################################
# class ReducedSpline1d:
#
#     def __init__(self):
#         pass
#
#     def autodiff(self, x, fn, n=4, args=None):
#         """ x ~ f
#             f ~ gwphase_3point5PN
#             n ~ 4 (spline degree + 1)
#             d = d/dx ~ d/df
#         """
#         ad = adipy.adn(x, n)
#         p = lib.fneval(ad, fn, args=args)
#         return p.d(n)
#
#     def delta(self, tol, deriv_n, n):
#         C3 = 5./384.
#         return (tol/C3/np.abs(deriv_n))**(1./n)


##############################################
class ReducedBasis:
    def __init__(self, inner, inner_type):
        self.inner = inner
        self.inner_type = inner_type

    def malloc(self, Nbasis, Npoints, Nquads, Nmodes=1, dtype="complex"):
        """Allocate memory for numpy arrays used for making reduced basis"""
        self.errors = lib.malloc("double", Nbasis)
        self.indices = lib.malloc("int", Nbasis)
        if Nmodes == 1:
            self.basis = lib.malloc(dtype, Nbasis, Nquads)
        elif Nmodes > 1:
            self.basis = lib.malloc(dtype, Nbasis, Nmodes, Nquads)
        else:
            raise Exception("Expected positive number of modes.")
        self.basisnorms = lib.malloc("double", Nbasis)
        self.alpha = lib.malloc(dtype, Nbasis, Npoints)

    def alpha1(self, e, h):
        """Inner product of a basis function e with a function h:
        alpha(e,h) = <e, h>
        """
        return self.inner.dot(e, h, self.inner_type)

    def alpha_arr(self, e, hs):
        """Inner products of a basis function e with an array of functions hs"""
        return np.array([self.alpha1(e, hh) for hh in hs])

    def proj_error_from_basis(self, basis, h):
        """Square of the projection error of a function h on basis"""
        norm = np.real(self.inner.norm(h, self.inner_type))
        dim = len(basis[:, 0])
        return norm**2 - np.sum(
            np.abs(self.alpha1(basis[ii], h)) ** 2 for ii in range(dim)
        )

    def proj_errors_from_basis(self, basis, hs):
        """Square of the projection error of functions hs on basis"""
        # norms = np.real([self.inner.norm(hh, self.inner_type) for hh in hs])
        # dim = len(basis[:,0])
        # return [norms**2-np.sum(np.abs(self.alpha1(basis[ii], hh))**2
        #         for ii in range(dim)) for hh in hs]
        return [self.proj_error_from_basis(basis, hh) for hh in hs]

    def proj_mismatch_from_basis(self, basis, h):
        """Mismatch of a function h with its projection onto the basis"""
        norms = np.real(self.inner.norm(h, self.inner_type))
        dim = len(basis[:, 0])
        return (
            1.0
            - np.real(
                np.sum(np.abs(self.alpha1(basis[ii], h)) ** 2 for ii in range(dim))
            )
            / norms
        )

    def proj_errors_from_alpha(self, alpha, norms=None):
        """Square of the projection error of a function h on basis
        in terms of pre-computed alpha matrix.
        """
        if norms is None:
            norms = np.ones(len(alpha[0]), dtype="double")
        else:
            norms = norms
        dim = len(alpha[:, 0])
        return norms**2 - np.sum(np.abs(alpha[ii]) ** 2 for ii in range(dim))

    def projection_from_basis(self, h, basis):
        """Project a function h onto the basis functions"""
        # return np.array([ee*self.alpha1(ee, h) for ee in basis])
        return np.sum(ee * self.alpha1(ee, h) for ee in basis)

    def projection_from_alpha(self, basis, alpha):
        """Project a function h onto the basis functions in terms of
        pre-computed alpha matrix entry.
        """
        return np.array([np.sum(basis[ii] * alpha[ii]) for ii in range(len(basis))])



from typing import NamedTuple

import lal
import lalsimulation
import numpy as np

from rombus.model import RombusModel


class model(RombusModel):

    model_dtype = complex

    params.add("m1", 30, 35)  # noqa F821
    params.add("m2", 30, 35)  # noqa F821
    params.add("chi1L", 0, 0.1)  # noqa F821
    params.add("chi2L", 0, 0.1)  # noqa F821
    params.add("chip", 0, 0.1)  # noqa F821
    params.add("thetaJ", -np.pi / 2, np.pi / 2)  # noqa F821
    params.add("alpha", 0, np.pi / 2)  # noqa F821
    params.set_validation(lambda p: p.m1 >= p.m2)  # noqa F821

    def cache(self):
        self.l1 = 0
        self.l2 = 0
        self.fmin = 20
        self.fmax = 1024
        self.deltaF = 1.0 / 4.0
        self.nf = int((self.fmax - self.fmin) / self.deltaF) + 1
        self.fseries = np.linspace(self.fmin, self.fmax, self.nf)
        self.fmin_index = int(self.fmin / self.deltaF)
        self.WFdict = lal.CreateDict()

    def set_domain(self):
        return self.fseries

    def compute(self, params: NamedTuple, domain) -> np.array:

        lalsimulation.SimInspiralWaveformParamsInsertTidalLambda1(self.WFdict, self.l1)
        lalsimulation.SimInspiralWaveformParamsInsertTidalLambda2(self.WFdict, self.l2)

        if not np.array_equiv(domain, self.fseries):
            h = lalsimulation.SimIMRPhenomPFrequencySequence(
                domain,
                params.chi1L,
                params.chi2L,
                params.chip,
                params.thetaJ,
                params.m1 * lal.lal.MSUN_SI,
                params.m2 * lal.lal.MSUN_SI,
                1e6 * lal.lal.PC_SI * 100,
                params.alpha,
                0,
                40,
                lalsimulation.IMRPhenomPv2NRTidal_V,
                lalsimulation.NRTidalv2_V,
                self.WFdict,
            )
            h = h[0].data.data
        else:
            h = lalsimulation.SimIMRPhenomP(
                params.chi1L,
                params.chi2L,
                params.chip,
                params.thetaJ,
                params.m1 * lal.lal.MSUN_SI,
                params.m2 * lal.lal.MSUN_SI,
                1e6 * lal.lal.PC_SI * 100,
                params.alpha,
                0,
                self.deltaF,
                self.fmin,
                self.fmax,
                40,
                lalsimulation.IMRPhenomPv2NRTidal_V,
                lalsimulation.NRTidalv2_V,
                self.WFdict,
            )
            h = h[0].data.data[self.fmin_index : len(h[0].data.data)]
            if len(h) < self.nf:
                h = np.append(h, np.zeros(self.nf - len(h), dtype=complex))

        return h



import importlib
import sys
import os
import shutil
import timeit
from abc import ABCMeta, abstractmethod
from collections import namedtuple, Counter
from typing import Any, List, Callable
from tqdm.auto import tqdm

import h5py
import numpy as np

import rombus._core.mpi as mpi

MAX_N_TRIES = 1000


class Params(object):
    def __init__(self):
        self.params = []
        self.names = []
        self.count = 0
        self._idx = -1
        self.params_dtype = None
        self._validate = lambda x: True

    def add(self, name, range_min, range_max):
        Param = namedtuple("Param", "name min max")
        self.params.append(Param(name, range_min, range_max))
        self.count = self.count + 1
        self.names.append(name)

        # Update the named tuple that will be used to convert numpy arrays to Param objects
        self.params_dtype = namedtuple("params_dtype", self.names)

    def set_validation(self, func: Callable):
        self._validate = func

    def generate_random_sample(self, random_generator):

        new_sample = np.ndarray(self.count, dtype=np.float64)
        n_tries = 0
        while True:
            for i, param in enumerate(self.params):
                new_sample[i] = random_generator.uniform(low=param.min, high=param.max)
            param = self.np2param(new_sample)
            if self._validate(param):
                break
            else:
                n_tries = n_tries + 1
                if n_tries >= MAX_N_TRIES:
                    raise Exception(
                        f"Max number of tries ({MAX_N_TRIES}) reached when trying to generate a valid random parameter set"
                    )
        return new_sample

    def np2param(self, params_np):
        return self.params_dtype(**dict(zip(self.names, np.atleast_1d(params_np))))

    def __iter__(self):
        self._idx = -1
        return self

    def __next__(self):
        self._idx = self._idx + 1
        if self._idx >= self.count:
            raise StopIteration
        return self.params[self._idx]


class _RombusModelMeta(type):
    def __prepare__(name, *args, **kwargs):
        """Initialise the dictionary that gets passed to __new___.

        This is needed here because we don't want the user to have to
        initialise the member(s) that we are adding.  This is the only
        method that gets sourced before the class code is executed, so
        it needs to be done here, not in __new___.
        """

        result = dict()

        # Initialise the following members↵
        result["params"] = Params()
        result["model_dtype"] = np.dtype("float64")

        return result

    def __new__(mcs, name, bases, dct):

        # Perform super-metaclass construction↵
        return super(_RombusModelMeta, mcs).__new__(mcs, name, bases, dct)


class _RombusModelABCMeta(_RombusModelMeta, ABCMeta):
    pass


class RombusModel(metaclass=_RombusModelABCMeta):
    def __init__(self, model_str):

        # Check that at least one parameter has beed defined
        assert self.params.count > 0

        # Ensure that model_dtype is of the right type
        # assert self.model_dtype.type == np.dtype

        # Run an optional init() method
        self.cache()

        # Initialise the domain
        self.domain = self.set_domain()
        self.n_domain = len(self.domain)

        # Check that the domain has been suitably set
        assert self.n_domain > 0

        # Keep track of the model string so we can reinstantiate from a saved state
        self.model_str = model_str
        self.model_basename = self.model_str.split(":")[0].split(".")[-1]

    def cache(self):
        pass

    @abstractmethod  # make sure this is the inner-most decorator
    def set_domain(self):
        pass

    @abstractmethod  # make sure this is the inner-most decorator
    def compute(self, params: np.ndarray, domain) -> np.ndarray:
        pass

    def generate_model_set(self, samples: List[np.ndarray]) -> np.ndarray:
        """returns a list of models (one for each row in 'samples')"""

        my_ts = np.zeros(
            shape=(samples.n_samples, self.n_domain), dtype=self.model_dtype
        )
        for i, params_numpy in enumerate(
            tqdm(samples.samples, desc=f"Generating training set for rank {mpi.RANK}")
        ):
            model_i = self.compute(self.params.np2param(params_numpy), self.domain)
            my_ts[i] = model_i / np.sqrt(np.vdot(model_i, model_i))

        return my_ts

    def parse_cli_params(self, args):
        """Parse parameters from arguments given on command line"""

        model_params = dict()
        for param_i in args:
            if not param_i.startswith("-"):
                res = param_i.split("=")
                if len(res) == 2:
                    # NOTE: for now, all parameters are assumed to be floats
                    model_params[res[0]] = float(res[1])
                else:
                    raise Exception(f"Don't know what to do with argument '{param_i}'")
            else:
                raise Exception(f"Don't know what to do with option '{param_i}'")

        # Check that all parameters are specified and that they match what is
        # defined in the model
        assert Counter(model_params.keys()) == Counter(self.params.names)

        return model_params

    def timing(self, samples):
        start_time = timeit.default_timer()
        for i, sample in enumerate(samples.samples):
            params_numpy = self.params.np2param(sample)
            _ = self.compute(params_numpy, self.domain)
        return timeit.default_timer() - start_time

    @classmethod
    def load(cls, model: str):
        """Import the model code"""

        model_class = import_from_string(model)
        return model_class(model)

    def write(self, h5file):
        """Save samples to file"""

        h5_group = h5file.create_group("model")
        h5_group.create_dataset("model_str", data=self.model_str)

    @classmethod
    def from_file(cls, file_in):
        """Create a ROM instance from a file"""

        close_file = False
        if not isinstance(file_in, str):
            h5file = file_in
        else:
            h5file = h5py.File(file_in, "r")
            close_file = True

        model_str = h5file["model/model_str"].asstr()[()]
        if close_file:
            h5file.close()
        return cls.load(model_str)

    @classmethod
    def write_project_template(cls, project_name):
        """Write a project model and sample file to start a new project from."""

        # Set the model we will template from
        model_name = "sinc"

        # Set source file paths
        pkgdir = sys.modules["rombus"].__path__[0]
        model_file_source = os.path.join(pkgdir, "models", f"{model_name}.py")
        samples_file_source = os.path.join(
            pkgdir, "models", f"{model_name}_samples.csv"
        )

        # Set output file paths
        model_file_out = os.path.join(os.getcwd(), f"{project_name}.py")
        samples_file_out = os.path.join(os.getcwd(), f"{project_name}_samples.csv")

        # Copy files
        shutil.copy(model_file_source, model_file_out)
        shutil.copy(samples_file_source, samples_file_out)


# The code that follows has been copied directly from the Uvicorn codebase: https://github.com/encode/uvicorn
# (commit: d613cbea388bafafb6f642077c035ed137deea61)
#
# Copyright © 2017-present, [Encode OSS Ltd](https://www.encode.io/).
# All rights reserved.


class ImportFromStringError(Exception):
    pass


def import_from_string(import_str: Any) -> Any:
    if not isinstance(import_str, str):
        return import_str

    module_str, _, attrs_str = import_str.partition(":")
    if not module_str or not attrs_str:
        message = (
            'Import string "{import_str}" must be in format "<module>:<attribute>".'
        )
        raise ImportFromStringError(message.format(import_str=import_str))

    try:
        module = importlib.import_module(module_str)
    except ImportError as exc:
        if exc.name != module_str:
            raise exc from None
        message = 'Could not import module "{module_str}".'
        raise ImportFromStringError(message.format(module_str=module_str))

    instance = module
    try:
        for attr_str in attrs_str.split("."):
            instance = getattr(instance, attr_str)
    except AttributeError:
        message = 'Attribute "{attrs_str}" not found in module "{module_str}".'
        raise ImportFromStringError(
            message.format(attrs_str=attrs_str, module_str=module_str)
        )

    return instance
